/*
	Copyright (c) 2005-2010, Akira Hayashi.
	All rights reserved.

	Redistribution and use in source and binary forms, with or
	without modification, are permitted provided that the following
	conditions are met:

	1. Redistributions of source code must retain the above copyright
	   notice, this list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above
	   copyright notice, this list of conditions and the following
	   disclaimer in the documentation and/or other materials provided
	   with the distribution.
	   
	3. Neither the name of the Akira Hayashi, RK Kaihatsu, nor the
	   names of its contributors may be used to endorse or promote
	   products derived from this software without specific prior
	   written permission.
	   
	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
	CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
	INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
	MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
	BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
	TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
	TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
	THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
	SUCH DAMAGE.
*/

#import "RKTextConverter.h"

static const unichar kCharacterLF = L'\n';
static const unichar kCharacterCR = L'\r';
static const unichar kCharacterLS = 0x2028;

@implementation RKTextConverter

- (id)init
{
	self = [super init];
	if( self )
	{
		_halfToFullMap = [[self makeHalfToFullMap] retain];
		_halfToFullDullnessMap = [[self makeHalfToFullDullnessMap] retain];
		_halfToFullSemiMap = [[self makeHalfToFullSemiMap] retain];
	}
	return self;
}

- (void)dealloc
{
	[_halfToFullMap release];
	[_halfToFullDullnessMap release];
	[_halfToFullSemiMap release];
	[super dealloc];
}

- (RKLineBreakType)lineBreakTypeOfString:(NSString*)str
{
	unichar			character;
	unsigned int	u, length;
	RKLineBreakType	lineBreakType = kRKLineBreakTypeUnknown;
	
	length = [str length];
	for( u = 0; u < length; u++ )
	{
		character = [str characterAtIndex:u];
		if( character == kCharacterCR )
		{
			if( u < (length - 1) && [str characterAtIndex:u+1] == kCharacterLF )
				lineBreakType = kRKLineBreakTypeCRLF;
			else
				lineBreakType = kRKLineBreakTypeCR;
		}
		else if( character == kCharacterLF )
			lineBreakType = kRKLineBreakTypeLF;
		else if( character == kCharacterLS )
			lineBreakType = kRKLineBreakTypeLS;
		
		if( lineBreakType != kRKLineBreakTypeUnknown )
			break;
	}
	return lineBreakType;
}

- (RKEncodingInfo)encodingInfoOfTextData:(NSData*)data
{
	RKEncodingDetectionWorkingStruct	workingStruct;
	BOOL	isDetected = NO;
	
	workingStruct.encodingInfo.encodingCode = kRKEncodingCodeUnknown;
	workingStruct.encodingInfo.byteOrder = CFByteOrderUnknown;
	workingStruct.encodingInfo.hasBOM = NO;
	workingStruct.encodingInfo.deviceCharMapping = kRKDeviceCharMappingMac;
	workingStruct.eucFeatures = 0;
	workingStruct.jisFeatures = 0;
	workingStruct.isNeverUTF8 = NO;
	workingStruct.textBytes = nil;
	workingStruct.curBytes = nil;
	workingStruct.textOffset = 0;
	workingStruct.textLength = 0;
	workingStruct.utf8Features = 0;
	
	if( !data )
		return workingStruct.encodingInfo;
	
	workingStruct.textBytes = [data bytes];
	workingStruct.curBytes = workingStruct.textBytes;
	workingStruct.textLength = [data length];
	
	if( [self detectEncodingInfoWithBOM:&workingStruct] )
		return workingStruct.encodingInfo;		//	OK
		
	for( ; workingStruct.textOffset < workingStruct.textLength; workingStruct.textOffset++, workingStruct.curBytes++ )
	{
		if( !workingStruct.isNeverUTF8 )
		{
			if( [self checkUTF8Features:&workingStruct] )
				continue;
		}
		
		if( [self checkUTFnFeatures:&workingStruct] )
		{
			isDetected = YES;
			break;
		}
		
		//	Check JIS features.
		if( *workingStruct.curBytes == 0x1B &&
			(workingStruct.textOffset + 2) < workingStruct.textLength )
		{
			if( *(workingStruct.curBytes + 1) == 0x24 &&
				( *(workingStruct.curBytes + 2) == 0x42 || *(workingStruct.curBytes + 2) == 0x40 ) )
			{
				workingStruct.encodingInfo.encodingCode = kRKEncodingCodeJIS;
				isDetected = YES;
				break;
			}
			else if( *(workingStruct.curBytes + 1) == 0x28 &&
					 ( *(workingStruct.curBytes + 2) == 0x4A || *(workingStruct.curBytes + 2) == 0x42 ) )
			{
				workingStruct.jisFeatures++;
				continue;
			}
		}
		
		// Check EUC, SJIS and UTF-8 features.
		if( (0x80 <= *workingStruct.curBytes && *workingStruct.curBytes <= 0x8D) ||
			(0x90 <= *workingStruct.curBytes && *workingStruct.curBytes <= 0x9F) )
		{
			// At this point, the text encoding may be the UTF-16 or
			// the UTF-32.
			while (workingStruct.textOffset < workingStruct.textLength)
			{
				if ([self checkUTFnFeatures:&workingStruct])
				{
					isDetected = YES;
					break;
				}
				workingStruct.textOffset++;
				workingStruct.curBytes++;
			}
			
			if (!isDetected)
			{
				// The text encoding is the SJIS or the UTF-8.
				workingStruct.encodingInfo.encodingCode = (workingStruct.utf8Features > 0) ? kRKEncodingCodeUTF8 : kRKEncodingCodeSJIS;
				isDetected = YES;
			}
			break;
		}
		else if( ![self isIncludedInASCII:*workingStruct.curBytes] &&
				 ![self isIncludedInJISX0201Roman:*workingStruct.curBytes] )
		{
			// Check the EUC features.
			if( *workingStruct.curBytes >= 0xA1 && *workingStruct.curBytes <= 0xFE )
			{
				// Check whether the character is a kanji.
				// (If the character is a kanji, both of the first byte
				// and the second byte are 0xA1 or greater and are
				// 0xFE or less.
				if( workingStruct.textOffset + 1 < workingStruct.textLength &&
					*(workingStruct.curBytes + 1) >= 0xA1 &&
					*(workingStruct.curBytes + 1) <= 0xFE )
				{
					workingStruct.eucFeatures++;
				}
			}
			else
			{
				workingStruct.eucFeatures++;
			}
		}
	}
	
	if( !isDetected )
	{
		if( workingStruct.jisFeatures == 0 &&
			workingStruct.eucFeatures == 0 )
		{
			workingStruct.encodingInfo.encodingCode = kRKEncodingCodeUnknown;
		}
		else if( workingStruct.utf8Features > workingStruct.jisFeatures &&
				 workingStruct.utf8Features > workingStruct.eucFeatures )
		{
			workingStruct.encodingInfo.encodingCode = kRKEncodingCodeUTF8;
		}
		else if( workingStruct.jisFeatures > 0 )
		{
			workingStruct.encodingInfo.encodingCode = kRKEncodingCodeJIS;
		}
		else if( workingStruct.eucFeatures == 0 )
		{
			workingStruct.encodingInfo.encodingCode = kRKEncodingCodeUnknown;
		}
		else
		{
			workingStruct.encodingInfo.encodingCode = kRKEncodingCodeEUCJP;
		}
	}
	return workingStruct.encodingInfo;
}

- (BOOL)detectEncodingInfoWithBOM:(RKEncodingDetectionWorkingStruct*)workingStruct
{
	BOOL	bRet = NO;
	
	if( workingStruct->textLength < 2 )
		return NO;
	
	if( workingStruct->textLength >= 3 &&
		workingStruct->textBytes[0] == 0xEF &&
		workingStruct->textBytes[1] == 0xBB &&
		workingStruct->textBytes[2] == 0xBF )
	{
		//	UTF8
		workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF8;
		workingStruct->encodingInfo.hasBOM = YES;
		bRet = YES;
	}
	else if( workingStruct->textBytes[0] == 0xFE &&
			 workingStruct->textBytes[1] == 0xFF )
	{
		//	UTF16 Big endian
		workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF16;
		workingStruct->encodingInfo.byteOrder = CFByteOrderBigEndian;
		workingStruct->encodingInfo.hasBOM = YES;
		bRet = YES;
	}
	else if( workingStruct->textBytes[0] == 0xFF &&
			 workingStruct->textBytes[1] == 0xFE )
	{
		workingStruct->encodingInfo.byteOrder = CFByteOrderLittleEndian;
		workingStruct->encodingInfo.hasBOM = YES;
		bRet = YES;

		if( workingStruct->textLength >= 4 &&
			workingStruct->textBytes[2] == 0x00 &&
			workingStruct->textBytes[3] == 0x00 )
		{
			//	UTF32 Little endian
			workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF32;
		}
		else
		{
			//	UTF16 Little endian
			workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF16;
		}
	}
	else if( workingStruct->textLength >= 4 &&
			 workingStruct->textBytes[0] == 0x00 &&
			 workingStruct->textBytes[1] == 0x00 &&
			 workingStruct->textBytes[2] == 0xFE &&
			 workingStruct->textBytes[3] == 0xFF )
	{
		//	UTF32 Big endian
		workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF32;
		workingStruct->encodingInfo.byteOrder = CFByteOrderBigEndian;
		workingStruct->encodingInfo.hasBOM = YES;
		bRet = YES;
	}
	
	return bRet;
}

- (BOOL)checkUTF8Features:(RKEncodingDetectionWorkingStruct*)workingStruct
{
	unsigned char	mask = 128;
	unsigned int	numOfOctets = 0;
	unsigned int	u;
	BOOL			bRet = NO;
	
	if( *workingStruct->curBytes == 0xFF ||
		*workingStruct->curBytes == 0xFE )
	{
		// In the UTF-8, 0xFE and 0xFF are never appeared in the octet.
		workingStruct->isNeverUTF8 = YES;
		workingStruct->utf8Features = 0;
	}
	else
	{
		for( u = 0; u < 6; u++ )
		{
			if( *workingStruct->curBytes & mask )
			{
				numOfOctets++;
				mask >>= 1;
			}
			else
			{
				break;
			}
		}
		
		if( numOfOctets > 1 )
		{
			for( u = 1; u < numOfOctets; u++ )
			{
				if( workingStruct->textOffset + u >= workingStruct->textLength ||
					(*(workingStruct->curBytes + u) & 128) == 0 ||
					(*(workingStruct->curBytes + u) & 64) != 0 )
				{
					break;
				}
			}
			
			if( u != numOfOctets ||
				workingStruct->textOffset + numOfOctets > workingStruct->textLength )
			{
				workingStruct->isNeverUTF8 = YES;
				workingStruct->utf8Features = 0;
			}
			else
			{
				workingStruct->utf8Features++;
				bRet = YES;
			}
		}
	}
	return bRet;
}
				
- (BOOL)checkUTFnFeatures:(RKEncodingDetectionWorkingStruct*)workingStruct
{
	unsigned int	nullCount = 0;
	unsigned int	u;
	BOOL			ret = NO;

	if( *workingStruct->curBytes == 0x00 )
		nullCount++;
	else
		return NO;
	
	for( u = 1; workingStruct->textOffset+u < workingStruct->textLength; u++ )
	{
		if( *(workingStruct->curBytes + u) == 0x00 )
			nullCount++;
		else
			break;
	}

	if (nullCount > 1)
	{
		// The text encoding is the UTF-32 because multiple NULL
		// characters are there.
		workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF32;
		if (workingStruct->textOffset % 4 == 0)
			workingStruct->encodingInfo.byteOrder = CFByteOrderBigEndian;
		else
			workingStruct->encodingInfo.byteOrder = CFByteOrderLittleEndian;
		ret = YES;
	}
	else if (nullCount == 1)
	{
		// UTF-16 or UTF-32
		if ((workingStruct->textOffset % 4) == 1 || (workingStruct->textOffset % 4) == 2 || (workingStruct->textLength % 4) != 0)
		{
			// UTF-16
			workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF16;
			if (workingStruct->textOffset % 2 == 0)
				workingStruct->encodingInfo.byteOrder = CFByteOrderBigEndian;
			else
				workingStruct->encodingInfo.byteOrder = CFByteOrderLittleEndian;
			ret = YES;
		}
		else if ((workingStruct->textOffset + 2) < workingStruct->textLength && (workingStruct->textOffset + 3) < workingStruct->textLength)
		{
			if (*(workingStruct->curBytes + 2) == 0 || *(workingStruct->curBytes + 3) == 0)
			{
				// UTF-16
				workingStruct->encodingInfo.encodingCode = kRKEncodingCodeUTF16;
				if (workingStruct->textOffset % 2 == 0)
					workingStruct->encodingInfo.byteOrder = CFByteOrderBigEndian;
				else
					workingStruct->encodingInfo.byteOrder = CFByteOrderLittleEndian;
				ret = YES;
			}
		}
	}
	
	return ret;
}

- (NSString*)stringWithData:(NSData*)textData encodingInfo:(RKEncodingInfo)encodingInfo
{
	NSStringEncoding		encoding;
	NSString*				str = nil;
	const unsigned char*	textBytes = [textData bytes];
	unsigned int			textLength = [textData length];
	BOOL					isValid = YES;
	
	if( textData )
	{
		if( encodingInfo.encodingCode == kRKEncodingCodeSJIS )
		{
			if (encodingInfo.deviceCharMapping == kRKDeviceCharMappingWindows)
				encoding = CFStringConvertEncodingToNSStringEncoding(kCFStringEncodingDOSJapanese);
			else
				encoding = CFStringConvertEncodingToNSStringEncoding(kCFStringEncodingMacJapanese);
		}
		else if( encodingInfo.encodingCode == kRKEncodingCodeJIS )
		{
			encoding = NSISO2022JPStringEncoding;
		}
		else if( encodingInfo.encodingCode == kRKEncodingCodeEUCJP )
		{
			encoding = NSJapaneseEUCStringEncoding;
		}
		else if( encodingInfo.encodingCode == kRKEncodingCodeUTF8 )
		{
			encoding = NSUTF8StringEncoding;
			if( encodingInfo.hasBOM )
			{
				textBytes += 3;
				textLength -= 3;
			}
		}
		else if( encodingInfo.encodingCode == kRKEncodingCodeUTF16 )
		{
			if( encodingInfo.hasBOM )
			{
				textBytes += 2;
				textLength -= 2;
			}
			
			if( encodingInfo.byteOrder == CFByteOrderBigEndian )
			{
				encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF16BE );
			}
			else if( encodingInfo.byteOrder == CFByteOrderLittleEndian )
			{
				encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF16LE );
			}
			else
			{
				isValid = NO;
			}
		}
		else if( encodingInfo.encodingCode == kRKEncodingCodeUTF32 )
		{
			if( encodingInfo.hasBOM )
			{
				textBytes += 4;
				textLength -= 4;
			}
			
			if( encodingInfo.byteOrder == CFByteOrderBigEndian )
			{
				encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF32BE );
			}
			else if( encodingInfo.byteOrder == CFByteOrderLittleEndian )
			{
				encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF32LE );
			}
			else
			{
				isValid = NO;
			}
		}
		else
		{
			encoding = NSMacOSRomanStringEncoding;
		}
		
		if (isValid)
			str = [[[NSString alloc] initWithBytes:textBytes length:textLength encoding:encoding] autorelease];
		
	}

	return str;
}

- (BOOL)isIncludedInASCII:(unsigned char)c
{
	return (c <= 0x20 || c == 0x7F);
}

- (BOOL)isIncludedInJISX0201Roman:(unsigned char)c
{
	return (0x21 <= c && c <= 0x7E);
}

- (void)convertLineBreakOfString:(NSMutableString*)string from:(RKLineBreakType)srcBreakType to:(RKLineBreakType)dstBreakType
{
	NSString*	srcBreakStr = [self lineBreakString:srcBreakType];
	NSString*	dstBreakStr = [self lineBreakString:dstBreakType];
	
	if( srcBreakType != dstBreakType )
	{
		[string replaceOccurrencesOfString:srcBreakStr withString:dstBreakStr options:0
			range:NSMakeRange(0, [string length])];
	}
}

- (NSString*)lineBreakString:(RKLineBreakType)breakType
{
	NSString*	str = nil;
	
	switch( breakType )
	{
		case kRKLineBreakTypeLF:
			str = [NSString stringWithFormat:@"%C", kCharacterLF];
			break;
		
		case kRKLineBreakTypeCR:
			str = [NSString stringWithFormat:@"%C", kCharacterCR];
			break;
			
		case kRKLineBreakTypeCRLF:
			str = [NSString stringWithFormat:@"%C%C", kCharacterCR, kCharacterLF];
			break;
			
		case kRKLineBreakTypeLS:
			str = [NSString stringWithFormat:@"%C", kCharacterLS];
			break;
			
	}
	return str;
}

- (NSData*)dataWithString:(NSString*)str encodingInfo:(RKEncodingInfo)encodingInfo
{
	NSMutableData*		tempData = nil;
	NSData*				data = nil;
	NSStringEncoding	encoding;
	
	if( !str )
		return nil;
	
	if( encodingInfo.encodingCode == kRKEncodingCodeSJIS )
	{
		if (encodingInfo.deviceCharMapping == kRKDeviceCharMappingWindows)
			encoding = CFStringConvertEncodingToNSStringEncoding(kCFStringEncodingDOSJapanese);
		else
			encoding = CFStringConvertEncodingToNSStringEncoding(kCFStringEncodingMacJapanese);
		
		data = [str dataUsingEncoding:encoding allowLossyConversion:NO];
		if (data == nil)
			[str dataUsingEncoding:encoding allowLossyConversion:YES];
	}
	else if( encodingInfo.encodingCode == kRKEncodingCodeJIS )
	{
		data = [str dataUsingEncoding:NSISO2022JPStringEncoding allowLossyConversion:NO];
		if (data == nil)
			data = [str dataUsingEncoding:NSISO2022JPStringEncoding allowLossyConversion:YES];
	}
	else if( encodingInfo.encodingCode == kRKEncodingCodeEUCJP )
	{
		data = [str dataUsingEncoding:NSJapaneseEUCStringEncoding allowLossyConversion:NO];
		if (data == nil)
			data = [str dataUsingEncoding:NSJapaneseEUCStringEncoding allowLossyConversion:YES];
	}
	else if( encodingInfo.encodingCode == kRKEncodingCodeUTF8 )
	{
		data = [str dataUsingEncoding:NSUTF8StringEncoding allowLossyConversion:NO];
		if (data == nil)
			data = [str dataUsingEncoding:NSUTF8StringEncoding allowLossyConversion:YES];
		
		if( encodingInfo.hasBOM )
		{
			unsigned char utf8bom[3] = { 0xEF, 0xBB, 0xBF };
			
			tempData = [NSMutableData dataWithCapacity:([data length] + 3)];
			[tempData appendBytes:utf8bom length:3];
			[tempData appendData:data];
			
			data = tempData;
		}
	}
	else if( encodingInfo.encodingCode == kRKEncodingCodeUTF16 &&
			 (encodingInfo.byteOrder == CFByteOrderBigEndian || encodingInfo.byteOrder == CFByteOrderLittleEndian) )
	{
		if( encodingInfo.byteOrder == CFByteOrderBigEndian )
			encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF16BE );
		else
			encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF16LE );
		
		data = [str dataUsingEncoding:encoding allowLossyConversion:NO];
		if (data == nil)
			data = [str dataUsingEncoding:encoding allowLossyConversion:YES];
		
		if( encodingInfo.hasBOM )
		{
			unsigned char utf16bom[2];
			
			if( encodingInfo.byteOrder == CFByteOrderBigEndian )
			{
				utf16bom[0] = 0xFE;
				utf16bom[1] = 0xFF;
			}
			else
			{
				utf16bom[1] = 0xFE;
				utf16bom[0] = 0xFF;
			}
			
			tempData = [NSMutableData dataWithCapacity:([data length] + 2)];
			[tempData appendBytes:utf16bom length:2];
			[tempData appendData:data];
			
			data = tempData;
		}
	}
	else if( encodingInfo.encodingCode == kRKEncodingCodeUTF32 &&
			 (encodingInfo.byteOrder == CFByteOrderBigEndian || encodingInfo.byteOrder == CFByteOrderLittleEndian) )
	{
		if( encodingInfo.byteOrder == CFByteOrderBigEndian )
			encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF32BE );
		else
			encoding = CFStringConvertEncodingToNSStringEncoding( kCFStringEncodingUTF32LE );
		
		data = [str dataUsingEncoding:encoding allowLossyConversion:NO];
		if (data == nil)
			data = [str dataUsingEncoding:encoding allowLossyConversion:YES];

		if( encodingInfo.hasBOM )
		{
			unsigned char utf32bom[4];

			if( encodingInfo.byteOrder == CFByteOrderBigEndian )
			{
				utf32bom[0] = 0x00;
				utf32bom[1] = 0x00;
				utf32bom[2] = 0xFE;
				utf32bom[3] = 0xFF;
			}
			else
			{
				utf32bom[3] = 0x00;
				utf32bom[2] = 0x00;
				utf32bom[1] = 0xFE;
				utf32bom[0] = 0xFF;
			}
			
			tempData = [NSMutableData dataWithCapacity:([data length] + 4)];
			[tempData appendBytes:utf32bom length:4];
			[tempData appendData:data];
			
			data = tempData;
		}
	}
	else
	{
		data = [str dataUsingEncoding:NSMacOSRomanStringEncoding allowLossyConversion:NO];
		if (data == nil)
			data = [str dataUsingEncoding:NSMacOSRomanStringEncoding allowLossyConversion:YES];
	}
	return data;
}

- (BOOL)isHalfwidthKatakana:(unichar)c
{
	return ( 0xFF61 <= c && c <= 0xFF9F );
}

- (unichar)convertHalfwidthToFullwidthKatakana:(unichar)srcChar
	nextChar:(unichar)nextChar numOfCharsBeReplaced:(unsigned int*)numOfCharsBeReplaced
{
	unichar		dstChar = srcChar;
	NSNumber*	num;
	
	if( *numOfCharsBeReplaced )
		*numOfCharsBeReplaced = 0;
	
	if( [self isHalfwidthKatakana:srcChar] )
	{
		if( nextChar == 0xFF9E || nextChar == 0xFF9F )
		{
			if( srcChar == 0xFF73 ||
				(0xFF76 <= srcChar && srcChar <= 0xFF84) ||
				(0xFF8A <= srcChar && srcChar <= 0xFF8E) )
			{
				if( nextChar == 0xFF9E )
					num = [_halfToFullDullnessMap objectForKey:[NSNumber numberWithUnsignedInt:srcChar]];
				else
					num = [_halfToFullSemiMap objectForKey:[NSNumber numberWithUnsignedInt:srcChar]];
				
				if( num )
					dstChar = [num unsignedIntValue];
				else
					NSLog(@"error");
					
				*numOfCharsBeReplaced = 2;
			}
		}
		
		if( srcChar == dstChar )
		{
			num = [_halfToFullMap objectForKey:[NSNumber numberWithUnsignedInt:srcChar]];
			if( num )
				dstChar = [num unsignedIntValue];

			*numOfCharsBeReplaced = 1;
		}
			
	}
	return dstChar;
}

- (NSDictionary*)makeHalfToFullMap
{
	NSMutableDictionary*	newMap = [NSMutableDictionary dictionaryWithCapacity:61];
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x3002] forKey:[NSNumber numberWithUnsignedInt:0xFF61]];	//。
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x300C] forKey:[NSNumber numberWithUnsignedInt:0xFF62]];	//「
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x300D] forKey:[NSNumber numberWithUnsignedInt:0xFF63]];	//」
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x3001] forKey:[NSNumber numberWithUnsignedInt:0xFF64]];	//、
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30FB] forKey:[NSNumber numberWithUnsignedInt:0xFF65]];	//・
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30F2] forKey:[NSNumber numberWithUnsignedInt:0xFF66]];	//ヲ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A1] forKey:[NSNumber numberWithUnsignedInt:0xFF67]];	//ァ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A3] forKey:[NSNumber numberWithUnsignedInt:0xFF68]];	//ィ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A5] forKey:[NSNumber numberWithUnsignedInt:0xFF69]];	//ゥ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A7] forKey:[NSNumber numberWithUnsignedInt:0xFF6A]];	//ェ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A9] forKey:[NSNumber numberWithUnsignedInt:0xFF6B]];	//ォ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E3] forKey:[NSNumber numberWithUnsignedInt:0xFF6C]];	//ャ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E5] forKey:[NSNumber numberWithUnsignedInt:0xFF6D]];	//ュ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E7] forKey:[NSNumber numberWithUnsignedInt:0xFF6E]];	//ョ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C3] forKey:[NSNumber numberWithUnsignedInt:0xFF6F]];	//ッ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30FC] forKey:[NSNumber numberWithUnsignedInt:0xFF70]];	//ー
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A2] forKey:[NSNumber numberWithUnsignedInt:0xFF71]];	//ア
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A4] forKey:[NSNumber numberWithUnsignedInt:0xFF72]];	//イ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A6] forKey:[NSNumber numberWithUnsignedInt:0xFF73]];	//ウ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30A8] forKey:[NSNumber numberWithUnsignedInt:0xFF74]];	//エ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30AA] forKey:[NSNumber numberWithUnsignedInt:0xFF75]];	//オ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30AB] forKey:[NSNumber numberWithUnsignedInt:0xFF76]];	//カ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30AD] forKey:[NSNumber numberWithUnsignedInt:0xFF77]];	//キ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30AF] forKey:[NSNumber numberWithUnsignedInt:0xFF78]];	//ク
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B1] forKey:[NSNumber numberWithUnsignedInt:0xFF79]];	//ケ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B3] forKey:[NSNumber numberWithUnsignedInt:0xFF7A]];	//コ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B5] forKey:[NSNumber numberWithUnsignedInt:0xFF7B]];	//サ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B7] forKey:[NSNumber numberWithUnsignedInt:0xFF7C]];	//シ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B9] forKey:[NSNumber numberWithUnsignedInt:0xFF7D]];	//ス
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30BB] forKey:[NSNumber numberWithUnsignedInt:0xFF7E]];	//セ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30BD] forKey:[NSNumber numberWithUnsignedInt:0xFF7F]];	//ソ
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30BF] forKey:[NSNumber numberWithUnsignedInt:0xFF80]];	//た
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C1] forKey:[NSNumber numberWithUnsignedInt:0xFF81]];	//チ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C4] forKey:[NSNumber numberWithUnsignedInt:0xFF82]];	//ツ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C6] forKey:[NSNumber numberWithUnsignedInt:0xFF83]];	//テ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C8] forKey:[NSNumber numberWithUnsignedInt:0xFF84]];	//ト
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30CA] forKey:[NSNumber numberWithUnsignedInt:0xFF85]];	//ナ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30CB] forKey:[NSNumber numberWithUnsignedInt:0xFF86]];	//ニ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30CC] forKey:[NSNumber numberWithUnsignedInt:0xFF87]];	//ヌ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30CD] forKey:[NSNumber numberWithUnsignedInt:0xFF88]];	//ネ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30CE] forKey:[NSNumber numberWithUnsignedInt:0xFF89]];	//ノ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30CF] forKey:[NSNumber numberWithUnsignedInt:0xFF8A]];	//ハ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D2] forKey:[NSNumber numberWithUnsignedInt:0xFF8B]];	//ヒ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D5] forKey:[NSNumber numberWithUnsignedInt:0xFF8C]];	//フ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D8] forKey:[NSNumber numberWithUnsignedInt:0xFF8D]];	//ヘ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30DB] forKey:[NSNumber numberWithUnsignedInt:0xFF8E]];	//ホ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30DE] forKey:[NSNumber numberWithUnsignedInt:0xFF8F]];	//マ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30DF] forKey:[NSNumber numberWithUnsignedInt:0xFF90]];	//ミ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E0] forKey:[NSNumber numberWithUnsignedInt:0xFF91]];	//ム
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E1] forKey:[NSNumber numberWithUnsignedInt:0xFF92]];	//メ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E2] forKey:[NSNumber numberWithUnsignedInt:0xFF93]];	//モ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E4] forKey:[NSNumber numberWithUnsignedInt:0xFF94]];	//ヤ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E6] forKey:[NSNumber numberWithUnsignedInt:0xFF95]];	//ユ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E8] forKey:[NSNumber numberWithUnsignedInt:0xFF96]];	//ヨ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30E9] forKey:[NSNumber numberWithUnsignedInt:0xFF97]];	//ラ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30EA] forKey:[NSNumber numberWithUnsignedInt:0xFF98]];	//リ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30EB] forKey:[NSNumber numberWithUnsignedInt:0xFF99]];	//ル
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30EC] forKey:[NSNumber numberWithUnsignedInt:0xFF9A]];	//レ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30ED] forKey:[NSNumber numberWithUnsignedInt:0xFF9B]];	//ロ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30EF] forKey:[NSNumber numberWithUnsignedInt:0xFF9C]];	//ワ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30F3] forKey:[NSNumber numberWithUnsignedInt:0xFF9D]];	//ン
	
	return newMap;
}

- (NSDictionary*)makeHalfToFullDullnessMap
{
	NSMutableDictionary*	newMap = [NSMutableDictionary dictionaryWithCapacity:19];
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30AC] forKey:[NSNumber numberWithUnsignedInt:0xFF76]];	//ガ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30AE] forKey:[NSNumber numberWithUnsignedInt:0xFF77]];	//ギ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B0] forKey:[NSNumber numberWithUnsignedInt:0xFF78]];	//グ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B2] forKey:[NSNumber numberWithUnsignedInt:0xFF79]];	//ゲ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B4] forKey:[NSNumber numberWithUnsignedInt:0xFF7A]];	//ゴ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B6] forKey:[NSNumber numberWithUnsignedInt:0xFF7B]];	//ザ
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30B8] forKey:[NSNumber numberWithUnsignedInt:0xFF7C]];	//ジ
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30BA] forKey:[NSNumber numberWithUnsignedInt:0xFF7D]];	//ズ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30BC] forKey:[NSNumber numberWithUnsignedInt:0xFF7E]];	//ゼ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30BE] forKey:[NSNumber numberWithUnsignedInt:0xFF7F]];	//ゾ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C0] forKey:[NSNumber numberWithUnsignedInt:0xFF80]];	//ダ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C2] forKey:[NSNumber numberWithUnsignedInt:0xFF81]];	//ヂ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C5] forKey:[NSNumber numberWithUnsignedInt:0xFF82]];	//ヅ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C7] forKey:[NSNumber numberWithUnsignedInt:0xFF83]];	//デ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30C9] forKey:[NSNumber numberWithUnsignedInt:0xFF84]];	//ド
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D0] forKey:[NSNumber numberWithUnsignedInt:0xFF8A]];	//バ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D3] forKey:[NSNumber numberWithUnsignedInt:0xFF8B]];	//ビ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D6] forKey:[NSNumber numberWithUnsignedInt:0xFF8C]];	//ブ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D9] forKey:[NSNumber numberWithUnsignedInt:0xFF8D]];	//ベ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30DC] forKey:[NSNumber numberWithUnsignedInt:0xFF8E]];	//ボ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30F4] forKey:[NSNumber numberWithUnsignedInt:0xFF73]];	//ヴ
	
	return newMap;
}

- (NSDictionary*)makeHalfToFullSemiMap
{
	NSMutableDictionary*	newMap = [NSMutableDictionary dictionaryWithCapacity:5];
	
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D1] forKey:[NSNumber numberWithUnsignedInt:0xFF8A]];	//パ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D4] forKey:[NSNumber numberWithUnsignedInt:0xFF8B]];	//ピ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30D7] forKey:[NSNumber numberWithUnsignedInt:0xFF8C]];	//プ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30DA] forKey:[NSNumber numberWithUnsignedInt:0xFF8D]];	//ペ
	[newMap setObject:[NSNumber numberWithUnsignedInt:0x30DD] forKey:[NSNumber numberWithUnsignedInt:0xFF8E]];	//ポ
	
	return newMap;
}

- (void)convertHalfToFullKatakanaOfString:(NSMutableString*)string
{
	unichar			curChar, nextChar, newChar;
	unsigned int	offset, length, numOfCharsBeReplaced;
	
	length = [string length];
	for( offset = 0; offset < length; offset++ )
	{
		NSAutoreleasePool*	pool = [[NSAutoreleasePool alloc] init];
		
		curChar = [string characterAtIndex:offset];
		
		if( [self isHalfwidthKatakana:curChar] )
		{
			nextChar = (offset < length - 1) ? [string characterAtIndex:offset+1] : 0;
			newChar = [self convertHalfwidthToFullwidthKatakana:curChar nextChar:nextChar numOfCharsBeReplaced:&numOfCharsBeReplaced];
			
			if( newChar != 0 )
			{
				[string replaceCharactersInRange:NSMakeRange( offset, numOfCharsBeReplaced )
					withString:[NSString stringWithFormat:@"%C", newChar]];
				length = [string length];
			}
		}
		
		[pool drain];
	}
}

		
@end
